///////////////////////////////////////////////
//
//	GemCollect by trapflag
//
//	GemCollect is an IDC script for the
//	Interactive Disassembler Pro by DataRescue
//	-------------------------------------------
//
//	Often you need to find the exact address of
//	a variable in a running process in memory.
//	These addresses are mostly stored in
//	dynamically allocated memory and may vary.
//	This script helps you to track down the
//	address(es) as well as the code accessing
//	those address(es).
//
//	Usage:
//	0.	Load the script into IDA.
//	1.	Run the process, then suspend it
//			and start a search for a known value.
//	2.	Resume the process, wait for the value
//			to change, then suspend the process
//			and continue the search for the now
//			changed value.
//	3.	Repeat step 2 until you found the
//			address holding your value.
//	4.	When you found the desired address,
//			you can set a hardware breakpoint on
//			it to reach the code modifying the
//			value.
//	-------------------------------------------
//
//	Because the script is an add-on for a very
//	powerful disassembler and a good debugger
//	in between, it has the following advantages
//	over similar standalone-tools:
//
//	* Pause the process anytime using IDA's
//		debugger
//	* Set (hardware) breakpoints on the
//		memory addresses found.
//	* Debug/reverse engineer the code accessing
//		the memory addresses.
//	* Use the script on a client computer while
//		running the process on a server using
//		IDA's remote debugging capabilities.
//
//		This is especially useful for processes
//		that prevent you to switch back to Windows!
//
//		For instructions on how to set up two
//		machines for remote debugging, please
//		refer to the excellent tutorials on
//		www.datarescue.com/idabase
//
//
//	If you want to purchase the Interactive
//	Disassembler (IDA) go to www.datarescue.com
//
//	For more plugins and IDC scripts visit
//	www.backtrace.de
//
//	-------------------------------------------
//
//	history:
//	--------
//	-	19.09.2004:
//		initial version.
//
//	-	25.09.2004:
//		added hotkey to re-print the addresses
//		found.
//
//	-	07.10.2004:
//		first public release.
//
//	For questions regarding this script:
//		trapflag(at)backtrace(dot)de
//
//	(c) 2004, trapflag
//
///////////////////////////////////////////////


#include <idc.idc>
#include "gemcollect.hpp"


static main(void)
{
	init();
	return;
}

static init(void)
{
	auto hotkey_start;
	auto hotkey_resume;
	auto hotkey_print;
	
	hotkey_start = "Shift-s";
	hotkey_resume = "Shift-r";
	hotkey_print = "Shift-p";

	Message("\nGemCollect v0.3b by tf\n"
			  "----------------------\n");
	
	reg_hotkey(hotkey_start,"start_search");
	reg_hotkey(hotkey_resume,"resume_search");
	reg_hotkey(hotkey_print,"print_results");
	Message("\n");
}

static reg_hotkey(hotkey, function)
{
	if (AddHotkey(hotkey,function) == IDCHK_OK)
	{
		Message("  -> %s to %s\n",hotkey,function);
	}
	else
	{
		Message("  -> ERROR: Hotkey %s could not be registered!\n",hotkey);
	}
}

static start_search(void)
{
	auto method;
	auto value;
	auto count;
	auto file;
	auto cfgfile;
	auto answer;
		
	cfgfile = open_file(get_cfgfile());
	method = get_cfg(cfgfile);
	if( method != BADADDR )
	{
		answer = AskYN(ID_CANCEL,"Really start a new search? Previous results will be lost!"); // -1:cancel,0-no,1-ok
		if( answer != ID_OK )
		{
			Message("Action cancelled!\n");
			return;
		}
		fclose(cfgfile);
		cfgfile = create_file(get_cfgfile());
	}
	method = ask_method();
	if(method == BADADDR)
	{
		Message("Unknown method! Aborting...\n");
		return;
	}
	writelong(cfgfile, method, 0);
	fclose(cfgfile);
	//file = create_file(get_tempfile());

	file = open_file(get_tempfile());
	
	value = AskLong(0,"Enter a value to search for:");
	Message("Search is now in progress, please wait...\n");
	count = save_addresses(file, value, method);
	Message("\nFound %d addresses.\n",count);
	fclose(file);
	return;
}

static resume_search(void)
{
	auto method;
	auto value;
	auto count;
	auto file;
	auto cfgfile;
	
	cfgfile = open_file(get_cfgfile());
	method = get_cfg(cfgfile);
	if( method == BADADDR )
	{
		method = ask_method();
		if(method == BADADDR)
		{
			Message("Unknown method! Aborting...\n");
			return;
		}
		fclose(cfgfile);
		cfgfile = create_file(get_cfgfile());
		writelong(cfgfile, method, 0);
		fclose(cfgfile);
	}	
	
	file = open_file(get_tempfile());
	value = AskLong(0,"Enter a value to search for:");
	Message("Search is now in progress, please wait...\n");
	count = compare_values(file, value, method);
	Message("\nFound %d addresses.\n",count);
	fclose(file);
	return;
}


static print_results(void)
{
	auto file;
	auto count;
	
	if( !file_exist(get_tempfile()) )
	{
		Message("There is no search in progress!\n");
		return 1;
	}
	file = open_file(get_tempfile());
	count = print_addresses(file);
	Message("\n%d addresses left.\n",count);
	fclose(file);
	return;
}

static print_addresses(file)
{
	auto ea;
	auto counter;
	
	while( !eof(file) )
	{
		ea = readlong(file, 0);
		if ( ea != BADADDR)
		{
			Message("%08X\n",ea);
			counter++;
		}
	}
	return counter;
}




static compare_values(file, value, method)
{
	auto ea;
	auto counter;
	
	while( !eof(file) )
	{
		ea = readlong(file, 0);
		if ( ea != BADADDR)
		{
			if( method == METHOD_DWORD )
			{
				if( Dword(ea) != value )
				{
					fseek(file, -4,1);
					writelong(file, BADADDR, 0);
				}
				else
				{
					Message("%08X\n",ea);
					counter++;
				}
			}
			else if( method == METHOD_WORD )
			{
				if( Word(ea) != value )
				{
					fseek(file, -4,1);
					writelong(file, BADADDR, 0);
				}
				else
				{
					Message("%08X\n",ea);
					counter++;
				}
			}
			else if( method == METHOD_BYTE )
			{
				if( Byte(ea) != value )
				{
					fseek(file, -4,1);
					writelong(file, BADADDR, 0);
				}
				else
				{
					Message("%08X\n",ea);
					counter++;
				}
			}
		}
	}
	return counter;
}

static save_addresses(file, value, method)
{
	auto ea;
	auto start_ea;
	auto end_ea;
	auto permissions;
	auto type;
	auto count;

	ea = FirstSeg();
	if (ea == BADADDR)
	{
		Message("No segments found!\n");
		return BADADDR;
	}
	do
	{
		type = GetSegmentAttr(ea,SEGATTR_TYPE);
		permissions = GetSegmentAttr(ea,SEGATTR_PERM);
		if( (permissions & 4 && permissions & 2) && type == 3)
		{
			start_ea = ea;
			end_ea = SegEnd(ea)-1;
			for(start_ea; start_ea <= end_ea-method; start_ea++)
			{
				if( method == METHOD_DWORD )
				{
					if(Dword(start_ea)==value)
					{
						Message("%08X\n",start_ea);
						writelong(file,start_ea,0);
						count++;
					}
				}
				else if( method == METHOD_WORD )
				{
					if(Word(start_ea)==value)
					{
						Message("%08X\n",start_ea);
						writelong(file,start_ea,0);
						count++;
					}
				}
				else if( method == METHOD_BYTE )
				{
					if( Byte(start_ea)==value )
					{
						Message("%08X\n",start_ea);
						writelong(file,start_ea,0);
						count++;
					}
				}
			}
		}
	} while( (ea=NextSeg(ea)) != BADADDR );
	return count;
}

static open_file(filename)
{
	auto file;
	
	if(!file_exist(filename))
	{
		file = fopen(filename,"wb");
	}
	else
	{
		file = fopen(filename,"rb+");
	}
	return file;
}

static create_file(filename)
{
	return fopen(filename,"wb+");
}

static eof(file)
{
	if( ftell(file) == filelength(file) ) return true;
	return false;
}


static file_exist(filename)
{
	auto fp;

	fp=fopen(filename,"rb");
	if(fp!=0)
	{
		fclose(fp);
		return 1;
	}
	return 0;
}

static get_tempfile(void)
{
	return GetIdaDirectory() + "\\gc_memdif.dat";
}

static get_cfgfile(void)
{
	return GetIdaDirectory() + "\\gc_cfg.dat";
}

static get_cfg(file)
{
	return readlong(file, 0);
}

static ask_method(void)
{
	auto method;
	
	method = AskLong(METHOD_DWORD, "Please enter a search method:\n"
								   "'4':  DWORD, '2': WORD, '1': BYTE\n");
	if((method!=1 && method !=2 && method !=4) || method == -1)
	{
		return BADADDR;
	}
	return method;
}